#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Name: RIFE ncnn Vulkan Python wrapper
Author: ArchieMeng
Date Created: March 29, 2021
Last Modified: May 21, 2021

Dev: K4YT3X
Last Modified: August 17, 2021
"""

# built-in imports
import importlib
import pathlib
import sys

# third-party imports
from PIL import Image

# local imports
if __package__ is None:
    import rife_ncnn_vulkan_wrapper as wrapped
else:
    wrapped = importlib.import_module(f"{__package__}.rife_ncnn_vulkan_wrapper")


class Rife:
    def __init__(
        self,
        gpuid: int = -1,
        model: str = "rife-v2.3",
        scale: int = 2,
        tta_mode: bool = False,
        uhd_mode: bool = False,
        num_threads: int = 1,
    ):
        # scale must be a power of 2
        if (scale & (scale - 1)) == 0:
            self.scale = scale
        else:
            raise ValueError("scale should be a power of 2")

        # determine if rife-v2 is used
        rife_v2 = ("rife-v2" in model) or ("rife-v3" in model)
        rife_v4 = "rife-v4" in model

        # create raw RIFE wrapper object
        self._rife_object = wrapped.RifeWrapped(
            gpuid, tta_mode, uhd_mode, num_threads, rife_v2, rife_v4
        )
        self._load(model)

    def _load(self, model: str, model_dir: pathlib.Path = None):

        # if model_dir is not specified
        if model_dir is None:
            model_dir = pathlib.Path(model)
            if not model_dir.is_absolute() and not model_dir.is_dir():
                model_dir = pathlib.Path(__file__).parent / "models" / model

        # if the model_dir is specified and exists
        if model_dir.exists():
            modeldir_str = wrapped.StringType()
            if sys.platform in ("win32", "cygwin"):
                modeldir_str.wstr = wrapped.new_wstr_p()
                wrapped.wstr_p_assign(modeldir_str.wstr, str(model_dir))
            else:
                modeldir_str.str = wrapped.new_str_p()
                wrapped.str_p_assign(modeldir_str.str, str(model_dir))

            self._rife_object.load(modeldir_str)

        # if no model_dir is specified but doesn't exist
        else:
            raise FileNotFoundError(f"{model_dir} not found")

    def process(self, image0: Image, image1: Image, timestep: float = 0.5) -> Image:
        # Return the image immediately instead of doing the copy in the upstream part which cause black output problems
        # The reason is that the upstream code use ncnn::Mat::operator=(const Mat& m) does a reference copy which won't
        # change our OutImage data.
        if timestep == 0.:
            return image0
        elif timestep == 1.:
            return image1

        image0_bytes = bytearray(image0.tobytes())
        image1_bytes = bytearray(image1.tobytes())
        channels = int(len(image0_bytes) / (image0.width * image0.height))
        output_bytes = bytearray(len(image0_bytes))

        # convert image bytes into ncnn::Mat Image
        raw_in_image0 = wrapped.Image(
            image0_bytes, image0.width, image0.height, channels
        )
        raw_in_image1 = wrapped.Image(
            image1_bytes, image1.width, image1.height, channels
        )
        raw_out_image = wrapped.Image(
            output_bytes, image0.width, image0.height, channels
        )

        self._rife_object.process(raw_in_image0, raw_in_image1, timestep, raw_out_image)
        return Image.frombytes(
            image0.mode, (image0.width, image0.height), bytes(output_bytes)
        )


class RIFE(Rife):
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
from pathlib import Path
from decimal import Decimal, getcontext

from PIL import Image, ImageChops, ImageStat
from rife_ncnn_vulkan_python import Rife, wrapped

tests_path = Path(__file__).parent
images_path = (
    tests_path.parent / "rife_ncnn_vulkan_python" / "rife-ncnn-vulkan" / "images"
)
# set Decimal precision
getcontext().prec = 4

# Use GPU 0 if available.
gpu_id = 0 if wrapped.get_gpu_count() > 0 else -1


def _calc_image_diff(image0: Image.Image, image1: Image.Image) -> float:
    """
    calculate the percentage of differences between two images

    :param image0 Image.Image: the first frame
    :param image1 Image.Image: the second frame
    :rtype float: the percent difference between the two images
    """
    difference = ImageChops.difference(image0, image1)
    difference_stat = ImageStat.Stat(difference)
    percent_diff = sum(difference_stat.mean) / (len(difference_stat.mean) * 255) * 100
    return percent_diff


def test_default():
    input_image0 = Image.open(images_path / "0.png")
    input_image1 = Image.open(images_path / "1.png")

    interpolator = Rife(gpu_id)
    output_image = interpolator.process(input_image0, input_image1)

    test_image = Image.open(tests_path / "0.5_default.png")
    percent_diff = _calc_image_diff(test_image, output_image)
    logging.getLogger().info(f"%diff: {percent_diff}")

    test_image.close()
    output_image.close()
    input_image0.close()
    input_image1.close()

    assert percent_diff < 0.5


def test_rife_v4():
    input_image0 = Image.open(images_path / "0.png")
    input_image1 = Image.open(images_path / "1.png")

    interpolator = Rife(gpu_id, model="rife-v4")
    output_image = interpolator.process(input_image0, input_image1)

    test_image = Image.open(tests_path / "0.5_v4.png")
    percent_diff = _calc_image_diff(test_image, output_image)
    logging.getLogger().info(f"%diff: {percent_diff}")

    input_image0.close()
    input_image1.close()
    output_image.close()
    test_image.close()

    assert percent_diff < 0.5


def test_rife_v4_timestep():
    input_image0 = Image.open(images_path / "0.png")
    input_image1 = Image.open(images_path / "1.png")

    interpolator = Rife(gpu_id, model="rife-v4")
    step = Decimal(0)
    while step <= 1:
        output_image = interpolator.process(input_image0, input_image1, float(step))
        test_image = Image.open(tests_path / f"{float(step)}_v4.png")
        percent_diff = _calc_image_diff(test_image, output_image)
        logging.getLogger().info(f"%diff: {percent_diff}")
        assert percent_diff < 0.5
        test_image.close()
        step += Decimal(0.2)

    input_image0.close()
    input_image1.close()



#!/usr/bin/python
# -*- coding: utf-8 -*-
import os
import pathlib
import setuptools
import shutil
import sys
import tempfile
import zipfile

# the directory that contains all RIFE models
# required only for building wheels
MODELS_PATH = (
    pathlib.Path(__file__).parent
    / "rife_ncnn_vulkan_python"
    / "rife-ncnn-vulkan"
    / "models"
)

# HACK: allows pip to load user site packages when running setup
if "PYTHONNOUSERSITE" in os.environ:
    os.environ.pop("PYTHONNOUSERSITE")

sys.path.append(
    str(
        pathlib.Path.home()
        / ".local/lib/python{}.{}/site-packages".format(
            sys.version_info.major, sys.version_info.minor
        )
    )
)
sys.path.append(
    "/usr/lib/python{}.{}/site-packages".format(
        sys.version_info.major, sys.version_info.minor
    )
)

# external modules must be imported after the hack
import cmake_build_extension
import requests


def download(url, chunk_size=4096):
    # create a temporary file to contain the downloaded file
    handle, path_str = tempfile.mkstemp()
    path = pathlib.Path(path_str)

    # create download stream and size
    stream = requests.get(url, stream=True, allow_redirects=True)
    total_size = int(stream.headers.get("content-length", 0))

    # print download information summary
    print(f"Downloading: {url}")
    print(f"Total size: {total_size}")
    print(f"Chunk size: {chunk_size}")
    print(f"Saving to: {path}")

    # Write content into file
    with os.fdopen(handle, "wb") as output:
        for chunk in stream.iter_content(chunk_size=chunk_size):
            output.write(chunk) if chunk else None

    # return the full path of saved file
    return path


def download_models() -> None:
    rife_ncnn_vulkan_zip = download(
        "https://github.com/nihui/rife-ncnn-vulkan/archive/"
        "c806e66490679aebc1b4a6832985e004fd552f46.zip"
    )

    with zipfile.ZipFile(rife_ncnn_vulkan_zip) as archive:
        with tempfile.TemporaryDirectory() as tempdir:
            archive.extractall(tempdir)
            shutil.move(
                pathlib.Path(tempdir) / archive.namelist()[0] / "models", MODELS_PATH
            )

    # remove the temporary files/dirs
    rife_ncnn_vulkan_zip.unlink()

cmake_flags = [
    "-DBUILD_SHARED_LIBS:BOOL=OFF",
    "-DCALL_FROM_SETUP_PY:BOOL=ON",
]
cmake_flags.extend(os.environ.get("CMAKE_FLAGS", "").split())

# when building bdist wheels/installing the package
if sys.argv[1] == "bdist_wheel":

    # if the models are not present
    # download and extract the models from nihui/rife-ncnn-vulkan
    if not MODELS_PATH.exists():
        download_models()


setuptools.setup(
    ext_modules=[
        cmake_build_extension.CMakeExtension(
            name="rife-ncnn-vulkan-python",
            install_prefix="rife_ncnn_vulkan_python",
            write_top_level_init="from .rife_ncnn_vulkan import Rife, RIFE, wrapped",
            source_dir=str(pathlib.Path(__file__).parent / "rife_ncnn_vulkan_python"),
            cmake_configure_options=cmake_flags,
        )
    ],
    cmdclass={"build_ext": cmake_build_extension.BuildExtension},
)



# Tencent is pleased to support the open source community by making ncnn available.
#
# Copyright (C) 2021 THL A29 Limited, a Tencent company. All rights reserved.
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import torch
import torch.nn as nn
import torch.nn.functional as F

class Model(nn.Module):
    def __init__(self):
        super(Model, self).__init__()

        self.attention_0_0 = nn.MultiheadAttention(embed_dim=64, num_heads=4)

        if torch.__version__ >= '1.9':
            self.attention_1_0 = nn.MultiheadAttention(embed_dim=40, num_heads=4, batch_first=True)

    def forward(self, x, y):
        x0, _ = self.attention_0_0(x, x, x)

        if torch.__version__ < '1.9':
            return x0

        y0, _ = self.attention_1_0(y, y, y)

        return x0, y0

def test():
    net = Model()
    net.eval()

    torch.manual_seed(0)
    x = torch.rand(1, 1, 64)
    y = torch.rand(1, 15, 40)

    a = net(x, y)

    # export torchscript
    mod = torch.jit.trace(net, (x, y))
    mod.save("test_nn_MultiheadAttention.pt")

    # torchscript to pnnx
    import os
    os.system("../../src/pnnx test_nn_MultiheadAttention.pt inputshape=[1,1,64],[1,15,40]")

    # ncnn inference
    import test_nn_MultiheadAttention_ncnn
    b = test_nn_MultiheadAttention_ncnn.test_inference()

    for a0, b0 in zip(a, b):
        if not torch.allclose(a0, b0, 1e-4, 1e-4):
            return False
    return True

if __name__ == "__main__":
    if test():
        exit(0)
    else:
        exit(1)
